;==========================================================

   lfsr64_rand PROTO range:DWORD
   lfsr_rand_vshift PROTO range:DWORD
   pdixon_lfsr64_rand PROTO range:DWORD
   pdixon_lfsr63_rand PROTO range:DWORD
   lc_rand PROTO range:DWORD
   init_crt_rand PROTO
   crt_rand PROTO range:DWORD
   csp_rand PROTO range:DWORD

;==========================================================
; This proc implements a 64-bit, maximum-length LFSR counter
; with XNOR feedback. The basic concept and feedback tap
; selections were taken from XILINX application note 052,
; July 7,1996 (Version 1.1). The feedback taps are from SR
; stages 60,61,63,64 (EAX bits 4,3,1,0). Rather than
; performing an actual XNOR operation on the tap bits, the
; code sets the feedback bit based on the parity of the tap
; bits (set for even, clear for odd). The counter performs
; the specified number of shifts per call and returns the
; count in EDX:EAX.
;
; For the original version of this procedure running on a
; P3, the clock count, including the call overhead, varied
; substantially from call to call, and averaged 640 for 64
; shifts per call. For this version, the clock count is a
; uniform 397 for 64 shifts, and 71 for 9 shifts.
;==========================================================
   .data
      shift_reg dd 3141592654,2718281828
   .code

align 4
lfsr64_counter proc shifts:DWORD
      push  ebx
      mov   eax,shift_reg
      mov   edx,shift_reg+4
      mov   ecx,shifts

   shiftLoop:

      ; Prevent partial register stall, saves
      ; 490 clocks per call with 64 shifts.
      xor   ebx,ebx

      ; Set high-order bit of EBX on
      ; PE, without forward jump.
      test  eax,1bh
      setpe bl
      shl   ebx,31

      shr   edx,1
      rcr   eax,1
      or    edx,ebx

      dec   ecx
      jnz   shiftLoop

      mov   shift_reg,eax
      mov   shift_reg+4,edx
      pop   ebx
      ret
lfsr64_counter endp
;==========================================================
; This proc returns a random 32-bit integer in the interval
; 0 to range - 1.
;
; The number of counter shifts per call is fixed at 64.
;==========================================================
lfsr64_rand proc range:DWORD
      push  64
      call  lfsr64_counter
      mul   range
      mov   eax,edx
      ret
lfsr64_rand endp
;==========================================================
; This proc returns a random 32-bit integer in the
; interval 0 to range - 1.
;
; This version uses variable shifts - the number of
; counter shifts per call is the number of significant
; bits in the range, plus 1.
;==========================================================
lfsr64_rand_vshift proc range:DWORD
      mov   eax,range
      xor   ecx,ecx
   @@:
      inc   ecx
      shr   eax,1
      jnz   @B
      inc   ecx
      push  ecx
      call lfsr64_counter
      mul   range
      mov   eax,edx
      ret
lfsr64_rand_vshift endp
;==========================================================
; This proc is pdixon's MMX version of the 64-bit LFSR
; counter, modified to return the counter value in EDX:EAX.
;
; Running on a P3, and including the call overhead, the
; clock count is 62.
;==========================================================
align 4
pdixon_lsfr64_counter proc
      ;
      movq  MM0,qword ptr shift_reg  ;get the last number
      movq  MM1,MM0        ;take a copy

      psrlq MM1,1          ;do 1st tap bit (LSBs)
      pxor  MM0,MM1
      psrlq MM1,2          ;do 2nd tap bit (LSBs)
      pxor  MM0,MM1
      psrlq MM1,1
      pxor  MM0,MM1        ;do 3rd tap bit (LSBs), low 60 bits now done
                           ;now correct for the high 4 bits
      movq  MM1,MM0        ;take a copy
      psllq MM1,60         ;correct 3rd tap MSBs
      pxor  MM0,MM1
      psllq MM1,1          ;correct 2nd tap MSBs
      pxor  MM0,MM1
      psllq MM1,2
      pxor  MM0,MM1        ;correct 1st tap MSBs, all done

      movq  qword ptr shift_reg,MM0   ;write back the result

      emms                 ;re-enable FPU

      mov   eax,shift_reg
      mov   edx,shift_reg+4

      ret
pdixon_lsfr64_counter endp
;==========================================================
; This proc returns a random 32-bit integer in the
; interval 0 to range - 1.
;==========================================================
pdixon_lfsr64_rand proc range:DWORD
      call  pdixon_lsfr64_counter
      mul   range
      mov   eax,edx
      ret
pdixon_lfsr64_rand endp
;==========================================================
; This proc is pdixon's MMX generator (based on a 63-bit
; LFSR?), modified to take a range argument and return
; the result in EAX.
;
; Running on a P3, and including the call overhead,
; the clock count is 58.
;==========================================================
   .data
      align 8
      rng      dq 1954678542154787h  ;a better seed
      mask2    dq 7fffffffffffffffh
   .code

align 4
pdixon_lfsr63_rand proc range:DWORD
      movq  MM0,rng  ;get the last number
      movq  MM1,MM0
      psllq MM0,1
      pxor  MM1,MM0
      pand  MM1,mask2
      movq  MM0,MM1
      psrlq MM1,62
      pxor  MM0,MM1
      movq  rng,MM0
      mov   eax,dword ptr rng ;get 32bits_of_random_number
      mul   range
      mov   eax,edx
      ret
pdixon_lfsr63_rand endp
;==========================================================
; This proc is a MASM version of the linear congruental
; generator from the HLA Standard Library, with a different
; constant and the result converted to a range.
;==========================================================
   .data
      lcSeed dd 0
   .code

lc_rand proc range:DWORD
      neg   lcSeed
      mov   eax,31415821
      mul   lcSeed
      sub   eax,edx
      jns   @F
      add   eax,2
   @@:
      dec   eax
      mov   lcSeed,eax
      mul   range
      mov   eax,edx
      ret
lc_rand endp
;==========================================================
; This proc loads msvcrt.dll and sets a pointer to
; the rand function.
;
; Returns zero for failure.
;==========================================================
   .data
      hLib        dd 0
      lpRandFunc  dd 0
   .code

init_crt_rand proc
      invoke LoadLibrary,SADD("msvcrt.dll")
      mov   hLib,eax
      or    eax,eax
      jnz   @F
      ret
   @@:
      invoke GetProcAddress,eax,SADD("rand")
      mov   lpRandFunc,eax
      ret
init_crt_rand endp
;==========================================================
; This proc calls the msvcrt.dll rand function and returns
; a random 32-bit integer in the interval 0 to range - 1.
;
; AFAIK rand is based on a linear congruental generator.
;
; Note that the rand function returns a value in the range
; 0 to RAND_MAX (= 7fffh).
;==========================================================
crt_rand proc range:DWORD
      mov   eax,lpRandFunc
      test  eax,eax
      jnz   @F
      ret
   @@:
      call  eax
      mov   ecx,range
      shl   ecx,1
      mul   cx
      movzx eax,dx
      ret
crt_rand endp
;==========================================================
; This proc collects random bytes generated by the
; cryptographic service provider (CSP), and returns a
; random 32-bit integer in the interval 0 to range - 1.
;
; Everything is in a single procedure to make automatic
; release of the CSP handle simple. The large buffer is
; a quick and dirty correction for the slowness of the
; CSP generator.
;==========================================================
    .data
    
hProv       dd  0
pNext       dd  0
      
    .data?      
    
rand_buffer dd  10000 dup(?)
      
    .code

PROV_RSA_FULL EQU 1
;==========================================================
csp_rand proc uses ebx range:DWORD

      mov   ebx,pNext
      .if ebx == 0
         invoke CryptAcquireContext,
                                    ADDR hProv,
                                    NULL,
                                    NULL,
                                    PROV_RSA_FULL,
                                    0
         .if eax == 0
            MsgBox NULL,"CryptAcquireContext failed",NULL,MB_OK
            mov selection, 0
            ret
         .endif

         invoke CryptGenRandom,hProv,10000*4,ADDR rand_buffer
         .if eax == 0
            MsgBox NULL,"CryptGenRandom failed",NULL,MB_OK
         .endif

         invoke CryptReleaseContext,hProv,0
      .endif

      mov   eax,[rand_buffer+ebx]
      add   ebx,4

      .if ebx > 9999*4
         xor   ebx,ebx
      .endif

      mov   pNext,ebx
      mul   range
      mov   eax,edx
      ret
csp_rand endp
;==========================================================
;  GetRadNumber(Range)
;----------------------------------------------------------
    .data

RandSeed    dd  1

    .code
;----------------------------------------------------------
CurrUse proc uses edx ecx   iRange

        mov ecx, 13
        xor eax, eax
_grnd:
        mov dx, word ptr RandSeed[2]
        and dh, 01001000b   ; 48h
        cmp dh, 00001000b   ; 08h
        jz  _xor1
        cmp dh, 01000000b   ; 40h
        jz  _xor1
        clc                 ; xor result=0 so clear carry
        jmp _shift
_xor1:
        stc                 ; xor result=1 so set carry
_shift:
        rcl RandSeed, 1
        rcl ax, 1
        dec ecx
        jnz _grnd

        mov eax, RandSeed
        mul iRange
        mov eax, edx
        ret

CurrUse endp
;==========================================================
; Ni+1 = (a*Ni + c) mod m
;
; Numerical research by S.K.Park and K.W.Miller has identified a
; theoretical "best" set of parameters.
;
; For the linear congruential algorithm to be effective, a and m can
; take on only a very few values, with m most certainly being prime.
;
; Park and Miller identified the parameter values a = 16807 (41A7h),
; m = 2147483647 (7fffffffh), c = 0 as producing the most statistically-
; random values for 32-bit signed integers.
;
; For producing 16-bit values, a good set of parameters is a = 171 (0ABh),
; m = 30269 (763Dh), c = 0.
;
; If an application requires 32-bit numbers, Park and Miller suggested
; a = 42871 (0A777h) or a = 69621 (10FF5h)."
;----------------------------------------------------------
    .data

ParkMiller_m    dd  7fffffffh
ParkMiller_N    dd  1
ParkMiller_a    dd  69621 ; 42871 ; 16807

    .code
;----------------------------------------------------------
ParkMiller  proc uses edx   iRange

    mov eax, ParkMiller_N   ; N(i+1) = a*N(i) mod m
    mul ParkMiller_a
  ; and eax, ParkMiller_m
    mov ParkMiller_N, eax

    mul iRange
    mov eax, edx

    ret

ParkMiller  endp
;==========================================================